from typing import List

from eis1600.nlp.cameltools import lemmatize_and_tag_ner


def annotate_miu_text(df):
    lemmas, ner_tags, pos_tags = ['_'], ['_'], ['_']
    section_id, temp_tokens = None, []
    for entry in list(zip(df['SECTIONS'].to_list(), df['TOKENS'].to_list()))[1:]:
        _section, _token = entry[0], entry[1]
        if _section is not None:
            # Start a new section
            if len(temp_tokens) > 0:
                # 1. process the previous section
                _labels = lemmatize_and_tag_ner(temp_tokens)
                _, _ner_tags, _lemmas, _dediac_lemmas, _pos_tags = zip(*_labels)
                ner_tags.extend(_ner_tags)
                lemmas.extend(_dediac_lemmas)
                pos_tags.extend(_pos_tags)

                # 2. reset variables
                section_id, temp_tokens = None, []

        token = _token if _token not in ['', None] else '_'
        temp_tokens.append(token)

    if len(temp_tokens) > 0:
        _labels = lemmatize_and_tag_ner(temp_tokens)
        _, _ner_tags, _lemmas, _dediac_lemmas, _pos_tags = zip(*_labels)
        ner_tags.extend(_ner_tags)
        lemmas.extend(_dediac_lemmas)
        pos_tags.extend(_pos_tags)
    return ner_tags, lemmas, pos_tags


def camel2md_as_list(labels: list) -> List[str]:
    default_str = ''
    types_mapping = {
        'LOC': 'ÜT',
        'PERS': 'ÜP'
    }
    converted_tokens, temp_tokens, temp_class = [], [], None
    for _label in labels:
        if _label is None:
            converted_tokens.append(default_str)
        else:
            # Check if the first letter of the label is 'o' or 'B' a Begining of an NE
            if _label[0] in ['O', 'B', '_']:
                if len(temp_tokens) > 0 and temp_class is not None:
                    converted_tokens.append(f"{types_mapping.get(temp_class, 'ÜM')}{len(temp_tokens)}")  # e.g. ÜP3
                    converted_tokens.extend([default_str] * (len(temp_tokens) - 1))
                    # reset temp variables
                    temp_tokens, temp_class = [], None

                if _label in ['O', '_']:
                    converted_tokens.append(default_str)
                else:
                    temp_tokens.append(_label)
                    temp_class = _label[2:]

            else:
                if temp_class is not None:
                    temp_tokens.append(_label)
                else:
                    converted_tokens.append(default_str)
    if len(temp_tokens) > 0 and temp_class is not None:
        converted_tokens.append(f"{types_mapping.get(temp_class, 'ÜM')}{len(temp_tokens)}")
        converted_tokens.extend([default_str] * (len(temp_tokens) - 1))
    return converted_tokens


from typing import Optional, Type, TextIO

from eis1600.miu.HeadingTracker import HeadingTracker
from eis1600.miu.YAMLHandler import YAMLHandler

from eis1600.markdown.re_patterns import MIU_HEADER_PATTERN, NEWLINES_CROWD_PATTERN


def create_yml_header(headings: Type[HeadingTracker]) -> str:
    """Creates a YAML header for the current MIU file and returns it as yamlfied string.

    :param Type[HeadingsTracker] headings: HeadingTracker with the super elements of the current MIU.
    :return str: YAML header for the current MIU.
    """
    yml_header = YAMLHandler()
    if headings:
        yml_header.set_headings(headings)

    return yml_header.get_yamlfied()


def extract_yml_header_and_text(miu_file_object: TextIO, is_header: Optional[bool] = False) -> (str, str):
    """ Returns the YAML header and the text as a tuple from MIU file object.

    Splits the MIU file into a tuple of YAML header and text.
    :param TextIO miu_file_object: File object of the MIU file from which to extract YAML header and text.
    :param bool is_header: Indicates if the current MIU is the YAML header of the whole work and if so skips
    removing
    blank lines, defaults to False.
    :return (str, str): Tuple of the extracted YAML header and text.
    """
    text = ''
    miu_yml_header = ''
    for line in iter(miu_file_object):
        if MIU_HEADER_PATTERN.match(line):
            # Omit the #MIU#Header# line as it is only needed inside the MIU.EIS1600 file, but not in YMLDATA.yml
            next(miu_file_object)
            line = next(miu_file_object)
            miu_yml_header = ''
            while not MIU_HEADER_PATTERN.match(line):
                miu_yml_header += line
                line = next(miu_file_object)
            # Omit the empty line between the header content and the #MIU#Header# line
            miu_yml_header = miu_yml_header[:-2]
            # Skip empty line after #MIU#Header#
            next(miu_file_object)
        else:
            text += line
        # Replace new lines which separate YAML header from text
        if not is_header:
            text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)

    return miu_yml_header, text


def update_yml_header():
    # TODO: update YAML header with MIU related information from automated analyses and manual tags
    pass



from camel_tools.tokenizers.word import simple_word_tokenize
from eis1600.miu.YAMLHandler import YAMLHandler
from pandas import DataFrame
from typing import List, Tuple

from openiti.helper.ara import normalize_ara_heavy

from eis1600.dates.Date import Date
from eis1600.dates.date_patterns import DATE_CATEGORIES, DATE_CATEGORIES_NOR, DATE_CATEGORY_PATTERN, DATE_PATTERN, \
    DAY_ONES_NOR, \
    DAY_TEN_NOR, MONTHS_NOR, \
    WEEKDAYS_NOR, ONES_NOR, TEN_NOR, HUNDRED_NOR
from eis1600.markdown.re_patterns import TAG_PATTERN


def tag_dates_headings(yml: YAMLHandler) -> YAMLHandler:
    headings = yml.headings
    for key, val in headings:
        if DATE_PATTERN.search(val):
            m = DATE_PATTERN.search(val)
            year = 0
            length = 1
            if m.group('ones'):
                year += ONES_NOR.get(normalize_ara_heavy(m.group('ones')))
                length += 1
            if m.group('ten'):
                year += TEN_NOR.get(normalize_ara_heavy(m.group('ten')))
                length += 1
            if m.group('hundred'):
                year += HUNDRED_NOR.get(normalize_ara_heavy(m.group('hundred')))
                length += 1
            yml.add_date(Date(year, length, 'H').get_tag())

    return yml


def tag_dates_fulltext(text: str) -> str:
    """Inserts date tags in the arabic text and returns the text with the tags.

    :param str text: arabic text.
    :return: arabic text with date tags.
    """
    text_updated = text
    m = DATE_PATTERN.search(text_updated)
    while m:
        month = None
        year = 0
        day = 0
        weekday = None
        # Length is one because sana is definitely recognized
        length = 1

        if DATE_CATEGORY_PATTERN.search(m.group('context')):
            last = DATE_CATEGORY_PATTERN.findall(m.group('context'))[-1]
            date_category = DATE_CATEGORIES_NOR.get(normalize_ara_heavy(last))
        else:
            date_category = 'X'

        # if m.group('weekday'):
        #     weekday = WEEKDAYS_NOR.get(normalize_ara_heavy(m.group('weekday')))
        # if m.group('day_ones'):
        #     day += DAY_ONES_NOR.get(normalize_ara_heavy(m.group('day_ones')))
        # if m.group('day_ten'):
        #     day += DAY_TEN_NOR.get(normalize_ara_heavy(m.group('day_ten')))
        # if m.group('month'):
        #     month_str = normalize_ara_heavy(m.group('month'))
        #     month = MONTHS_NOR.get(month_str)
        # else:
        #     mm = MONTH_PATTERN.search(m[0])
        #     if mm:
        #         month_str = mm[0]
        #         month = MONTHS.get(month_str)
        if m.group('ones'):
            year += ONES_NOR.get(normalize_ara_heavy(m.group('ones')))
            length += 1
        if m.group('ten'):
            year += TEN_NOR.get(normalize_ara_heavy(m.group('ten')))
            length += 1
        if m.group('hundred'):
            year += HUNDRED_NOR.get(normalize_ara_heavy(m.group('hundred')))
            length += 1

        # if day == 0:
        #     day = None
        if year == 0:
            year = None

        date = Date(year, length, date_category)
        pos = m.start('sana')
        text_updated = text_updated[:pos] + date.get_tag() + text_updated[pos:]

        m = DATE_PATTERN.search(text_updated, m.end('sana') + len(date.get_tag()))

    return text_updated


def date_annotate_miu_text(ner_df: DataFrame, yml: YAMLHandler) -> Tuple[List, YAMLHandler]:
    """Annotate dates in the MIU text, returns a list of tag per token.

    :param DataFrame ner_df: df containing the 'TOKENS' column.
    :param YAMLHandler yml: yml_header to collect date tags in.
    :return Tuple[List, YAMLHandler]: List of date tags per token, which can be added as additional column to the df,
    and modified yml.
    """
    ner_df.mask(ner_df == '', None, inplace=True)
    tokens = ner_df['TOKENS'].dropna()
    ar_text = ' '.join(tokens)

    tagged_text = tag_dates_fulltext(ar_text)
    tokens = simple_word_tokenize(tagged_text)
    ar_tokens, tags = [], []
    tag = None
    for t in tokens:
        if TAG_PATTERN.match(t):
            tag = t
            yml.add_date(t)
        else:
            ar_tokens.append(t)
            tags.append(tag)
            tag = None

    ner_df.loc[ner_df['TOKENS'].notna(), 'DATE_TAGS'] = tags

    return ner_df['DATE_TAGS'].fillna('').tolist(), yml



from pathlib import Path

import sys
import os
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from glob import glob
from multiprocessing import Pool

from eis1600.helper.repo import get_files_from_eis1600_dir, write_to_readme, read_files_from_readme
from eis1600.markdown.methods import insert_uids


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and os.path.isfile(input_arg):
            filepath, fileext = os.path.splitext(input_arg)
            if fileext != '.EIS1600TMP':
                parser.error('You need to input an EIS1600TMP file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to insert UIDs in EIS1600TMP file(s) and thereby converting them to final EIS1600 
            file(s).
-----
Give a single EIS1600TMP file as input
or 
Give an input AND an output directory for batch processing.

Use -e <EIS1600_repo> to batch process all EIS1600TMP files in the EIS1600 directory which have not been processed yet.
'''
            )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            '-e', '--eis1600_repo', type=str,
            help='Takes a path to the EIS1600 file repo and batch processes all files which have not been processed yet'
            )
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600TMP file to process, you need to run this command from inside text repo',
            action=CheckFileEndingAction
            )
    arg_parser.add_argument(
            'output', type=str, nargs='?',
            help='Optional, if given batch processes all files from the input directory to the output directory'
            )
    args = arg_parser.parse_args()

    verbose = args.verbose

    # If this script is run with a single file as input
    if args.input and not args.output:
        infile = './' + args.input
        if 'data' in infile:
            path = infile.split('data')[0]
        else:
            depth = len(infile.split('/'))
            if depth == 2:
                path = '../../../'
            elif depth == 3:
                path = '../../'
            else:
                path = '../'
        print(f'Insert UIDs into {infile}')
        insert_uids(infile, None, verbose)
        infiles = [infile.split('/')[-1]]
        write_to_readme(path, infiles, '# Texts converted into `.EIS1600`\n', '.EIS1600', True)

    # If this script is run with input and output directory (old, not used any more)
    elif args.output:
        input_dir = args.input
        output_dir = args.output

        print(f'Insert UIDs into {input_dir}, save resulting EIS1600 files to {output_dir}')

        infiles = glob(input_dir + '/*.EIS1600TMP')
        if not infiles:
            print(
                    'The input directory does not contain any EIS1600TMP files to process'
                    )
            sys.exit()

        # Check if output directory exists else create that directory
        Path(output_dir).mkdir(exist_ok=True, parents=True)

        params = [(infile, output_dir, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(insert_uids, params).get()

    # If this script is run with -e option on the EIS1600 text repo
    elif args.eis1600_repo:
        input_dir = args.eis1600_repo
        if not input_dir[-1] == '/':
            input_dir += '/'

        # Get EIS1600TMP files which are marked as ready to process in the README and have not yet been converted to
        # EIS1600
        print(f'Insert UIDs into files from the EIS1600 repo (only if there is not an EIS1600 file yet)')
        files_list = read_files_from_readme(input_dir, '# Texts converted into `.EIS1600TMP`\n')
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600TMP', 'EIS1600')
        if not infiles:
            print(
                    'There are no more EIS1600TMP files to process'
                    )
            sys.exit()

        # Insert UIDs in the selected files, runs in parallel
        params = [(infile, None, verbose) for infile in infiles]
        with Pool() as p:
            p.starmap_async(insert_uids, params).get()

        # Write list of processed files to the README
        write_to_readme(input_dir, infiles, '# Texts converted into `.EIS1600`\n', '.EIS1600', True)
    else:
        print('Pass in a <uri.EIS1600TMP> file to process a single file or use the -e option for batch processing')
        sys.exit()

    print('Done')
